@@ -167,33 +167,29 @@ err:
 static int hc_cli_io_handler(struct appctx *appctx)
 {
 	struct stream_interface *si = appctx->owner;
-	struct buffer *trash = alloc_trash_chunk();
 	struct httpclient *hc = appctx->ctx.cli.p0;
 	struct http_hdr *hdrs, *hdr;
 
-	if (!trash)
-		goto out;
 	if (appctx->ctx.cli.i0 & HC_CLI_F_RES_STLINE) {
-		chunk_appendf(trash, "%.*s %d %.*s\n", (unsigned int)istlen(hc->res.vsn), istptr(hc->res.vsn),
+		chunk_printf(&trash, "%.*s %d %.*s\n", (unsigned int)istlen(hc->res.vsn), istptr(hc->res.vsn),
 			      hc->res.status, (unsigned int)istlen(hc->res.reason), istptr(hc->res.reason));
-		if (ci_putchk(si_ic(si), trash) == -1)
-			si_rx_room_blk(si);
+		if (ci_putchk(si_ic(si), &trash) == -1)
+			goto more;
 		appctx->ctx.cli.i0 &= ~HC_CLI_F_RES_STLINE;
-		goto out;
 	}
 
 	if (appctx->ctx.cli.i0 & HC_CLI_F_RES_HDR) {
+		chunk_reset(&trash);
 		hdrs = hc->res.hdrs;
 		for (hdr = hdrs; isttest(hdr->v); hdr++) {
-			if (!h1_format_htx_hdr(hdr->n, hdr->v, trash))
-				goto out;
+			if (!h1_format_htx_hdr(hdr->n, hdr->v, &trash))
+				goto too_many_hdrs;
 		}
-		if (!chunk_memcat(trash, "\r\n", 2))
-			goto out;
-		if (ci_putchk(si_ic(si), trash) == -1)
-			si_rx_room_blk(si);
+		if (!chunk_memcat(&trash, "\r\n", 2))
+			goto too_many_hdrs;
+		if (ci_putchk(si_ic(si), &trash) == -1)
+			goto more;
 		appctx->ctx.cli.i0 &= ~HC_CLI_F_RES_HDR;
-		goto out;
 	}
 
 	if (appctx->ctx.cli.i0 & HC_CLI_F_RES_BODY) {
@@ -202,27 +198,28 @@ static int hc_cli_io_handler(struct appctx *appctx)
 		ret = httpclient_res_xfer(hc, &si_ic(si)->buf);
 		channel_add_input(si_ic(si), ret); /* forward what we put in the buffer channel */
 
-		if (!httpclient_data(hc)) {/* remove the flag if the buffer was emptied */
-			appctx->ctx.cli.i0 &= ~HC_CLI_F_RES_BODY;
-		}
-		goto out;
+		/* remove the flag if the buffer was emptied */
+		if (httpclient_data(hc))
+			goto more;
+		appctx->ctx.cli.i0 &= ~HC_CLI_F_RES_BODY;
 	}
 
 	/* we must close only if F_END is the last flag */
 	if (appctx->ctx.cli.i0 ==  HC_CLI_F_RES_END) {
-		si_shutw(si);
-		si_shutr(si);
 		appctx->ctx.cli.i0 &= ~HC_CLI_F_RES_END;
-		goto out;
+		goto end;
 	}
 
-out:
-	/* we didn't clear every flags, we should come back to finish things */
-	if (appctx->ctx.cli.i0)
-		si_rx_room_blk(si);
-
-	free_trash_chunk(trash);
+more:
+	si_rx_room_blk(si);
+	if (!appctx->ctx.cli.i0)
+		si_rx_endp_done(si);
 	return 0;
+end:
+	return 1;
+
+too_many_hdrs:
+	return cli_err(appctx, "Too many headers.\n");
 }
 
 static void hc_cli_release(struct appctx *appctx)
@@ -595,12 +592,10 @@ static void httpclient_applet_io_handler(struct appctx *appctx)
 	struct htx_blk *blk = NULL;
 	struct htx *htx;
 	struct htx_sl *sl = NULL;
-	int32_t pos;
 	uint32_t hdr_num;
 	uint32_t sz;
 	int ret;
 
-
 	while (1) {
 
 		/* required to stop */
@@ -726,35 +721,26 @@ static void httpclient_applet_io_handler(struct appctx *appctx)
 						goto more;
 
 					hdr_num = 0;
-
-					for (pos = htx_get_head(htx);  pos != -1; pos = htx_get_next(htx, pos)) {
-						struct htx_blk *blk = htx_get_blk(htx, pos);
+					blk = htx_get_head_blk(htx);
+					while (blk) {
 						enum htx_blk_type type = htx_get_blk_type(blk);
 						uint32_t sz = htx_get_blksz(blk);
 
-						if (type == HTX_BLK_UNUSED) {
-							c_rew(res, sz);
-							htx_remove_blk(htx, blk);
-						}
+						c_rew(res, sz);
 
 						if (type == HTX_BLK_HDR) {
 							hdrs[hdr_num].n = istdup(htx_get_blk_name(htx, blk));
 							hdrs[hdr_num].v = istdup(htx_get_blk_value(htx, blk));
-							if (!isttest(hdrs[hdr_num].v) || !isttest(hdrs[hdr_num].n))
-								goto end;
-							c_rew(res, sz);
-							htx_remove_blk(htx, blk);
 							hdr_num++;
 						}
-
-						/* create a NULL end of array and leave the loop */
-						if (type == HTX_BLK_EOH) {
+						else if (type == HTX_BLK_EOH) {
+							/* create a NULL end of array and leave the loop */
 							hdrs[hdr_num].n = IST_NULL;
 							hdrs[hdr_num].v = IST_NULL;
-							c_rew(res, sz);
 							htx_remove_blk(htx, blk);
 							break;
 						}
+						blk = htx_remove_blk(htx, blk);
 					}
 
 					if (hdr_num) {
@@ -798,8 +784,8 @@ static void httpclient_applet_io_handler(struct appctx *appctx)
 					goto process_data;
 
 				/* decapsule the htx data to raw data */
-				for (pos = htx_get_head(htx); pos != -1; pos = htx_get_next(htx, pos)) {
-					struct htx_blk *blk = htx_get_blk(htx, pos);
+				blk = htx_get_head_blk(htx);
+				while (blk) {
 					enum htx_blk_type type = htx_get_blk_type(blk);
 					size_t count = co_data(res);
 					uint32_t blksz = htx_get_blksz(blk);
@@ -821,7 +807,7 @@ static void httpclient_applet_io_handler(struct appctx *appctx)
 						c_rew(res, vlen);
 
 						if (vlen == blksz)
-							htx_remove_blk(htx, blk);
+							blk = htx_remove_blk(htx, blk);
 						else
 							htx_cut_data_blk(htx, blk, vlen);
 
@@ -838,7 +824,7 @@ static void httpclient_applet_io_handler(struct appctx *appctx)
 
 						/* remove any block which is not a data block */
 						c_rew(res, blksz);
-						htx_remove_blk(htx, blk);
+						blk = htx_remove_blk(htx, blk);
 					}
 				}
 
